///////////////////////////////////////////////////////////
// Inserting edges into triangulations.                  //
// Two classes:                                          //
//                                                       //
//     ETraingulation   Represents triangulation of a    //
//                      set of ponts.                    //
//     EApplet          The front end.                   //
//                                                       //
// Sergei Savchenko (savs@cs.mcgill.ca)                  //
///////////////////////////////////////////////////////////

import java.awt.*;
import java.lang.Math;

///////////////////////////////////////////////////////////
// Representation of a triangulation of a set of points. //
///////////////////////////////////////////////////////////

class ETriangulation
{
 public String status;
 int size;                                  // allocated array size
 boolean show_num;
 boolean open_edge_del;

 int markx,marky;

 int nodes_x[],nodes_y[];                   // coordinates of nodes
 int no_nodes;

 int triangles_v1[],triangles_v2[],triangles_v3[];
 int no_triangles;

 int hull_v[];                              // indeces of hull vertices.
 int no_hull;

 int hole[];
 int no_hole;

 ETriangulation(int sz)
 {
  status="";                                // status line
  size=sz;
  markx=marky=-1;
  show_num=false;
  open_edge_del=false;
  no_nodes=no_triangles=no_hull=no_hole=0;

  nodes_x=new int[size];                    // allocate space
  nodes_y=new int[size];
  triangles_v1=new int[size];
  triangles_v2=new int[size];
  triangles_v3=new int[size];
  hull_v=new int[size];
  hole=new int[size];
 }

 void empty()
 {
  no_nodes=no_triangles=no_hull=0;
 }

 void show(Graphics g)                      // draws the triangulation
 {
  int i,cx,cy;

  for(i=0;i<no_triangles;i++)               // draw all triangles
  {
   g.setColor(Color.black);
   g.drawLine(nodes_x[triangles_v1[i]],nodes_y[triangles_v1[i]],
              nodes_x[triangles_v2[i]],nodes_y[triangles_v2[i]]
             );
   g.drawLine(nodes_x[triangles_v2[i]],nodes_y[triangles_v2[i]],
              nodes_x[triangles_v3[i]],nodes_y[triangles_v3[i]]
             );
   g.drawLine(nodes_x[triangles_v3[i]],nodes_y[triangles_v3[i]],
              nodes_x[triangles_v1[i]],nodes_y[triangles_v1[i]]
             );
  }

  g.setColor(Color.black);
  for(i=0;i<no_nodes;i++)                   // draw all nodes
  {
   g.drawRect(nodes_x[i]-2,nodes_y[i]-2,5,5);
   if(show_num)
    g.drawString(" "+i,nodes_x[i]+2,nodes_y[i]);
  }

  g.setColor(Color.red);                    // draw the convex hull
  for(i=0;i<no_hull;i++)
  {
   g.drawRect(nodes_x[hull_v[i]]-3,nodes_y[hull_v[i]]-3,7,7);
   if(show_num)
    g.drawString(" "+i,nodes_x[hull_v[i]]-13,nodes_y[hull_v[i]]);
  }

  g.setColor(Color.blue);                   // show a marker
  if((markx!=-1)&&(marky!=-1))
   g.drawRect(markx-4,marky-4,9,9);
 }

 double distance(int v1,int v2)             // distance from vertex to a point
 {
  int dx=nodes_x[v1]-nodes_x[v2];
  int dy=nodes_y[v1]-nodes_y[v2];
  return(Math.sqrt(dx*dx+dy*dy));
 }

 double distance(int v1,int x,int y)        // distance from vertex to a point
 {
  int dx=nodes_x[v1]-x;
  int dy=nodes_y[v1]-y;
  return(Math.sqrt(dx*dx+dy*dy));
 }

 double area(int v1,int v2,int v3)          // area covered by verteces
 {
  return((nodes_x[v1]*(nodes_y[v2]-nodes_y[v3])+
          nodes_x[v2]*(nodes_y[v3]-nodes_y[v1])+
          nodes_x[v3]*(nodes_y[v1]-nodes_y[v2]))/2);
 }

 int signArea(int v1,int v2,int v3)         // sign of a triangle's area
 {
  int sa=(nodes_x[v1]*(nodes_y[v2]-nodes_y[v3])+
          nodes_x[v2]*(nodes_y[v3]-nodes_y[v1])+
          nodes_x[v3]*(nodes_y[v1]-nodes_y[v2]))/2;
  if(sa==0) return(0);
  else if(sa>0) return(1);
       else return(-1);
 }

 boolean edgeIntersect(int v1,int v2,int v3,int v4)
 {
  int s1=signArea(v1,v2,v3);
  int s2=signArea(v1,v2,v4);
  int s3=signArea(v3,v4,v1);
  int s4=signArea(v3,v4,v2);

  if((v1==v3)||(v1==v4)||(v2==v3)||(v2==v4)) return(false);
  if(s1==s2) return(false);
  if(s3==s4) return(false);
  return(true);
 }

 int insideTriangle(int v,int t)
 {
  int a1,a2,a3,posit=signArea(triangles_v1[t],triangles_v2[t],triangles_v3[t]);

  a1=signArea(triangles_v1[t],triangles_v2[t],v);
  a2=signArea(triangles_v2[t],triangles_v3[t],v);
  a3=signArea(triangles_v3[t],triangles_v1[t],v);
  if((a1==posit)&&(a2==posit)&&(a3==posit))
  {
   return(1);                               // inside the triangle
  }
  else
  {
   return(0);                               // outside this triangle
  }
 }

 int hasVertex(int t,int v)
 {
  if(triangles_v1[t]==v) return(triangles_v2[t]);
  if(triangles_v2[t]==v) return(triangles_v3[t]);
  if(triangles_v3[t]==v) return(triangles_v1[t]);
  return(-1);
 }

 int o1,o2;
 void otherVerteces(int t,int v)
 {
  if(triangles_v1[t]==v) { o1=triangles_v2[t]; o2=triangles_v3[t]; }
  if(triangles_v2[t]==v) { o1=triangles_v1[t]; o2=triangles_v3[t]; }
  if(triangles_v3[t]==v) { o1=triangles_v1[t]; o2=triangles_v2[t]; }
 }

 int hasVerteces(int t,int v1,int v2)
 {
  if(triangles_v1[t]==v1)
  {
   if(triangles_v2[t]==v2) return(triangles_v3[t]);
   if(triangles_v3[t]==v2) return(triangles_v2[t]);
  }
  if(triangles_v2[t]==v1)
  {
   if(triangles_v1[t]==v2) return(triangles_v3[t]);
   if(triangles_v3[t]==v2) return(triangles_v1[t]);
  }
  if(triangles_v3[t]==v1)
  {
   if(triangles_v2[t]==v2) return(triangles_v1[t]);
   if(triangles_v1[t]==v2) return(triangles_v2[t]);
  }
  return(-1);
 }

 void addTriangle(int v1,int v2,int v3)
 {
  triangles_v1[no_triangles]=v1;
  triangles_v2[no_triangles]=v2;
  triangles_v3[no_triangles]=v3;

  no_triangles++;
 }

 void deleteTriangle(int t)                 // discard a triangle
 {
  if(no_triangles==1) no_triangles=0;
  else
  {
   no_triangles--;
   triangles_v1[t]=triangles_v1[no_triangles];
   triangles_v2[t]=triangles_v2[no_triangles];
   triangles_v3[t]=triangles_v3[no_triangles];
  }
 }

 void createEdgeHole(int v,int e)            // convex hull or Steiner pts
 {
  int t;

  hole[0]=e;
  no_hole=1;
  for(t=0;t<no_triangles;t++)
   if(hasVerteces(t,v,e)!=-1)
   {
    hole[no_hole]=hasVerteces(t,v,e);
    no_hole++;
    deleteTriangle(t);
    break;
   }

  for(boolean sm=true;sm;)
   for(sm=false,t=0;t<no_triangles;t++)
    if(hasVerteces(t,v,hole[no_hole-1])!=-1)
    {
     hole[no_hole]=hasVerteces(t,v,hole[no_hole-1]);
     no_hole++;
     deleteTriangle(t);
     sm=true;
     break;
    }
 }

 void createStarHole(int v)                 // internal vertex
 {
  int t;

  no_hole=0;
  for(t=0;t<no_triangles;t++)
   if(hasVertex(t,v)!=-1)
   {
    hole[no_hole]=hasVertex(t,v);
    no_hole++;
    deleteTriangle(t);
    break;
   }

  for(boolean sm=true;sm;)
   for(sm=false,t=0;t<no_triangles;t++)
    if(hasVerteces(t,v,hole[no_hole-1])!=-1)
    {
     hole[no_hole]=hasVerteces(t,v,hole[no_hole-1]);
     no_hole++;
     deleteTriangle(t);
     sm=true;
     break;
    }
 }

 int triangulateEV(int hole[],int n)
 {
  int k,i;

  hole[n]=hole[0];
  for(k=0;hole[k+2]!=hole[0];)
  {
   if(signArea(hole[k],hole[k+1],hole[k+2])==1) { k++; }
   else
   {
    addTriangle(hole[k],hole[k+1],hole[k+2]);
    for(i=k+1;i<n;i++) hole[i]=hole[i+1]; n--;
    if(k!=0) k--;
   }
  }
  return(n);
 }

 void steiner(int v,int k,int e1,int e2)
 {
  nodes_x[no_nodes]=(nodes_x[e1]+nodes_x[e2])/2;
  nodes_y[no_nodes]=(nodes_y[e1]+nodes_y[e2])/2;

  while(distance(e1,e2)>1)
  {
   nodes_x[no_nodes]=(nodes_x[e1]+nodes_x[e2])/2;
   nodes_y[no_nodes]=(nodes_y[e1]+nodes_y[e2])/2;
   if(area(v,k,no_nodes)>0) e1=no_nodes;
   else e2=no_nodes;
  }
  no_nodes++;
 }

 void triangulateSS(int hole[],int n,int kernel)
 {
  int i,e;
  int left[]=new int[size];
  int right[]=new int[size];
  int no_left=0,no_right=0;

  if(signArea(hole[0],kernel,hole[1])!=1)
  {
   for(i=0;i<n/2;i++) { int t=hole[i]; hole[i]=hole[n-i-1]; hole[n-i-1]=t; }
  }

  left[no_left]=hole[no_left]; no_left++;
  for(i=1;i<n;i++)
   if(signArea(hole[0],kernel,hole[i])==1) { left[no_left]=hole[i]; no_left++; }
   else break;
  steiner(hole[0],kernel,hole[i-1],hole[i]);
  left[no_left]=no_nodes-1; no_left++;

  right[no_right]=no_nodes-1; no_right++;
  for(;i<n;i++) { right[no_right]=hole[i]; no_right++; }
  right[no_right]=hole[0]; no_right++;
  e=right[1];

  triangulateEV(left,no_left);
  triangulateEV(right,no_right);            // triangulate the two parts

  createEdgeHole(no_nodes-1,e);             // remove Steiner point
  no_nodes--;

  triangulateEV(hole,no_hole);
  removeVertex(kernel);                     // remove kernel vertex
 }

 void insertNode(int x,int y)               // insert a point and update
 {
  int i;

  nodes_x[no_nodes]=x;
  nodes_y[no_nodes]=y;

  if(no_nodes==2)
  {
   addTriangle(0,1,2);                      // first triangle
   if(signArea(0,1,2)==1)
   {                                        // assume triangle is convex
    hull_v[0]=0; hull_v[1]=1; hull_v[2]=2;
   }
   else
   {
    hull_v[0]=2; hull_v[1]=1; hull_v[2]=0;  // make sure hull counterclockwise
   }
   no_hull=3;
  }
  else if(no_nodes>2)
  {
   boolean outside=true;

   for(i=0;i<no_triangles;i++)              // attempt inserting into
   {                                        // some triangle
    if(insideTriangle(no_nodes,i)==1)
    {
     addTriangle(triangles_v1[i],triangles_v2[i],no_nodes);
     addTriangle(triangles_v2[i],triangles_v3[i],no_nodes);
     addTriangle(triangles_v3[i],triangles_v1[i],no_nodes);
     deleteTriangle(i);
     outside=false;
     break;
    }
   }
   if(outside)                              // attempt connecting to
   {                                        // the convex hull
    int newhull_v[]=new int[size];          // hull's likely to change
    int no_newhull=0;                       // no points yet

    for(i=0;i<no_hull;i++)                  // connect to convex hull
    {
     int s1=signArea(hull_v[(i+no_hull-1)%no_hull],hull_v[i%no_hull],no_nodes);
     int s2=signArea(hull_v[i%no_hull],hull_v[(i+1)%no_hull],no_nodes);

     if(s1==1)
     {
      if(s2==1) {                           // not joined
                 newhull_v[no_newhull]=hull_v[i%no_hull]; no_newhull++;
                }
      else      {                           // beginning of the join
                 newhull_v[no_newhull]=hull_v[i%no_hull]; no_newhull++;
                 newhull_v[no_newhull]=no_nodes; no_newhull++;

                 addTriangle(hull_v[i%no_hull],hull_v[(i+1)%no_hull],no_nodes);
                }
     }
     else
     {
      if(s2==1) {                           // end of join
                 newhull_v[no_newhull]=hull_v[i%no_hull]; no_newhull++;
                }
      else      {                           // the join
                 addTriangle(hull_v[i%no_hull],hull_v[(i+1)%no_hull],no_nodes);
                }
     }
    }
    hull_v=newhull_v;
    no_hull=no_newhull;
   }
  }
  no_nodes++;
 }

 void removeVertex(int v)
 {
  no_nodes--;
  nodes_x[v]=nodes_x[no_nodes];
  nodes_y[v]=nodes_y[no_nodes];

  for(int i=0;i<no_triangles;i++)
  {
   if(triangles_v1[i]==no_nodes) triangles_v1[i]=v;
   if(triangles_v2[i]==no_nodes) triangles_v2[i]=v;
   if(triangles_v3[i]==no_nodes) triangles_v3[i]=v;
  }
  for(int i=0;i<no_hull;i++)
   if(hull_v[i]==no_nodes) hull_v[i]=v;
 }

 boolean deleteNode(int x,int y)
 {
  int v,t,i,j;

  for(v=0;v<no_nodes;v++)
  {
   if(distance(v,x,y)<4)
   {
    if(no_triangles<2) return(true);
    for(i=0;i<no_hull;i++) if(hull_v[i]==v) break;

    if(i<no_hull)                           // deleting an convex hull vertex
    {
     int newhull[]=new int[size];
     int no_newhull=0;
     createEdgeHole(v,hull_v[(i+no_hull-1)%no_hull]);
     no_hole=triangulateEV(hole,no_hole);

     for(j=0;j<i;j++) { newhull[no_newhull]=hull_v[j]; no_newhull++; }
     for(j=1;j<no_hole-1;j++) { newhull[no_newhull]=hole[j]; no_newhull++; }
     for(j=i+1;j<no_hull;j++) { newhull[no_newhull]=hull_v[j]; no_newhull++; }

     hull_v=newhull;
     no_hull=no_newhull;
     removeVertex(v);
    }
    else                                    // deleting an internal vertex
    {
     createStarHole(v);
     triangulateSS(hole,no_hole,v);
    }
    return(true);
   }
  }
  return(false);
 }

 void plugEdge(int v1,int v2)
 {
  int t,i,v=0;
  int left[]=new int[size];
  int right[]=new int[size];
  int no_left=0;
  int no_right=0;

  right[no_right]=left[no_left]=v1;
  no_left++; no_right++;

  for(t=0;t<no_triangles;t++)
  {
   if(hasVertex(t,v1)!=-1)
   {
    otherVerteces(t,v1);
    if(edgeIntersect(v1,v2,o1,o2)) break;
   }
  }
  if(t==no_triangles) return;

  deleteTriangle(t);
  for(;;)
  {
   if(signArea(v1,v2,o1)!=1) { int tmp=o1; o1=o2; o2=tmp; }
   if(right[no_right-1]!=o1) { right[no_right]=o1; no_right++; }
   if(left[no_left-1]!=o2) { left[no_left]=o2; no_left++; }

   for(i=0;i<no_triangles;i++)
    if((v=hasVerteces(i,o1,o2))!=-1) break;
   deleteTriangle(i);

   if(edgeIntersect(v1,v2,v,o1)) { o2=v;}
   else if(edgeIntersect(v1,v2,v,o2)) { o1=v; }
   else break;
  }
  right[no_right]=left[no_left]=v2;
  no_right++; no_left++;

  for(i=0;i<no_left/2;i++)                  // reverse the order
  { int tmp=left[i]; left[i]=left[no_left-1-i]; left[no_left-1-i]=tmp; }

  triangulateEV(right,no_right);
  triangulateEV(left,no_left);              // triangulate the two parts
 }

 void insertEdge(int firstx,int firsty,int x,int y)
 {
  int v,t,n1=0,n2=0;

  for(v=0;v<no_nodes;v++)
   if(distance(v,firstx,firsty)<4) { n1=v; break; }
  if(v==no_nodes) insertNode(firstx,firsty);
  for(v=0;v<no_nodes;v++)
   if(distance(v,x,y)<4) { n2=v; break; }
  if(v==no_nodes) insertNode(x,y);          // insert the verteces

  for(v=0;v<no_nodes;v++)
  {
   if((nodes_x[v]==firstx)&&(nodes_y[v]==firsty)) n1=v;
   if((nodes_x[v]==x)&&(nodes_y[v]==y)) n2=v;
  }

  plugEdge(n1,n2);                          // insert the edge itself
 }

 boolean deleteEdge(int firstx,int firsty,int x,int y)
 {
  int n1=-1,n2=-1,v,t,i,j;

  for(v=0;v<no_nodes;v++)
  {
   if(distance(v,firstx,firsty)<7) n1=v;
   if(distance(v,x,y)<4) n2=v;
  }
  if((n1==-1)||(n2==-1)) return(false);

  for(t=0;t<no_triangles;t++)
  {
   if(hasVerteces(t,n1,n2)!=-1)
   {
    if(open_edge_del)                       // find a stubbing edge
    {
     for(i=0;i<no_nodes;i++)
      for(j=0;j<no_nodes;j++)
       if(i!=j) if(edgeIntersect(n1,n2,i,j))
       {
        plugEdge(i,j); return(true);
       }
    }
    else                                    // delete the end-points
    {
     deleteNode(firstx,firsty);
     deleteNode(x,y);
    }
    return(true);
   }
  }
  return(false);
 }
}

///////////////////////////////////////////////////////////
// The front end.                                        //
///////////////////////////////////////////////////////////

public class EApplet extends java.applet.Applet
{
 ETriangulation tri;
 Image offscreen;
 Graphics offgraphics;
 Dimension offscreensize;
 boolean first;
 int firstx,firsty;

 public void init()
 {
  first=true;
  tri=new ETriangulation(512);
  add(new Checkbox("Show indices",null,false));
  add(new Checkbox("Use open edge deletion",null,false));
  add(new Button("Clear"));
 }

 public void paint(Graphics g) { repaint(); }

 public void update(Graphics g)
 {
  Dimension d=size();
  if((offscreen==null)||(d.width!=offscreensize.width)||(d.height!=offscreensize.height))
  {
   offscreen=createImage(d.width,d.height);
   offscreensize=d;
   offgraphics=offscreen.getGraphics();
   offgraphics.setFont(getFont());
  }
  offgraphics.setColor(Color.white);
  offgraphics.fillRect(0,0,d.width,d.height);
  offgraphics.setColor(Color.black);
  tri.show(offgraphics);
  offgraphics.drawString(tri.status,5,35);
  g.drawImage(offscreen,0,0,null);
 }

 public boolean action(Event evt,Object arg)
 {
  if(arg instanceof Boolean)
  {
   if(((Checkbox)evt.target).getLabel().equals("Show indices"))
    tri.show_num=((Boolean)arg).booleanValue();
   else
    tri.open_edge_del=((Boolean)arg).booleanValue();
   repaint();
   return true;
  }
  if("Clear".equals(arg))
  {
   tri.empty();
   repaint();
   return true;
  }
  return false;
 }

 public boolean mouseDown(Event evt,int x,int y)
 {
  if(first)
  {
   tri.markx=x; tri.marky=y;
   firstx=x; firsty=y; first=false;
  }
  else
  {
   tri.markx=-1; tri.marky=-1;
   first=true;
   if(!tri.deleteEdge(firstx,firsty,x,y))
    tri.insertEdge(firstx,firsty,x,y);
  }
  repaint();
  return(true);
 }
}

///////////////////////////////////////////////////////////


